/* ---------------------------------------------------------------------
	File:          FLUX.C - Core code				97-08-28
	Type:          `C' function
	Description:
		For nuclear reactor design and analysis.  Calculates core
		steady state flux given a 2-D grid and cell data for up to
		4 neutron groups.  Handles variable grid sizes, space dependent D
		and sigmas, and arbitrary neutron sources.
  ---------------------------------------------------------------------
*/
#include <math.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <time.h>
#include <process.h>
#include <string.h>

/* required for fscanf -- file input  */
#include <io.h>
#include <fcntl.h>
#include <sys\types.h>
#include <sys\stat.h>

int G; /* number of energy groups */
int num_cel; /* number of cells */
struct cel_db
{
	char name[5];            /* cell id */
	float *sigma_a;          /* absorption x-sect for each cell */
	float *sigma_f;          /* fission x-sect for each cell */
	float *nusigma_f;        /* nufission x-sect for each cell */
	float *chi;              /* distribution of fission neutrons */
	float *S_ext;            /* external source of fission neutrons */
	float *sigma_tr;         /* transport x-sect for each cell */
	float *sigma_r;          /* removal x_sect for each cell */
	float **sigma_s;         /* scattering x-sect for each cell */
} *cel;
typedef struct cel_db CELLIB;

int num_col;    /* number of columns of cells */
int max_col;
float *dx;      /* grid segment size (column width) */
int num_row;    /* number of rows of cells */
int max_row;
float *dy;      /* grid segment size (row width) */
float del_ext;  /* flux extrapolation length */
struct grd_db
{
	char name[5];           /* grid point name */
	float *sigma_a;         /* absorption X-sect for each grid point */
	float *nusigma_f;       /* nufission X-sect for each grid point */
	float *chi;             /* distribution of fission neutrons */
	float *S_ext;           /* external source of fission neutrons */
	float *sigma_r;         /* removal X-sect for each grid point */
	float *dif;             /* diffusion coefficient each grid point */
	float *flux;            /* neutron flux for each grid point */
	float *flux_old;        /* old neutron flux for each grid point */
	float **sigma_s;        /* scattering X-sect for each grid point */
} **grd;
typedef struct grd_db GRDLIB;

#define FALSE 0
#define TRUE  1
#define UNKNOWN 2
#define USER 3
#define JUST_TESTING 0

struct ctl_db
{
	int max_iter_flux;      /* maximum number of iterations in solution of Ax=b */
	float tolerance_flux;   /* convergence criterion */
	int max_iter_k;
	float tolerance_k;
	float w;                /* modification parameter for SOR  1<w<2 */
	float k1;               /* criticality parameter for SOR */
	float k2;               /* criticality parameter for SOR */
	float a;                /* gain for proportional term */
	float b;                /* gain for differential term */
} ctl;

float t_init, t_fin, dt, dt_init, dt_min, dt_max, dt_print;
float flx_setpoint, *v;

/* file names */
char cel_filename[13], grd_filename[13], out_filename[13], hist_filename[13];

/* initialize flags */
int dim = 0;    /* Flags dimension of selected model */
int cel_set = 0;
int grd_set = 0;
int flx_set = 0;
float flx_profile = 0.0;
int flx_norm = 0;
int flx_tran = 0;
int error_code = 0;

/* Function prototypes */

int cell_input(char *filename);
int grid_input(char *filename);
int flux_input(float flx_profile);
int SORM_loop (float w);
float SORM_step (float w, int n);
int storefluxm(void);
int storetrans(float t);
int tran_loop (void);
float tran_step (int g, float dt);
void post_message(char *message);

/* Constants */
#define PI      3.141593

/* -------------------------main routine---------------------------------- */
int main(int argc, char *argv[])
{
  /* Initialize */
  int err_flag = 0;

  strcpy(cel_filename, "cell.out");
  strcpy(grd_filename, "flux.inp");
  strcpy(out_filename, "flux.out");
  if (argc > 1 && strlen(argv[1]) < 13) strcpy (cel_filename, argv[1]);
  if (argc > 2 && strlen(argv[2]) < 13) strcpy (grd_filename, argv[2]);
  if (argc > 3 && strlen(argv[3]) < 13) strcpy (out_filename, argv[3]);

  /* Initialize data sets */
  cel_set = cell_input(cel_filename);
  if (cel_set != 1)
  {
		printf("filename %s invalid", cel_filename);
		exit(0);
  }

  grd_set = grid_input (grd_filename);
  if (grd_set != 1)
  {
		printf("filename %s invalid", grd_filename);
		exit(0);
  }
  flx_set = flux_input(flx_profile);
  if (flx_set != 1)
  {
		printf("unable to initialize flux profile type %d", flx_profile);
		exit(0);
  }
  err_flag = SORM_loop (ctl.w);
  if (err_flag == -1)printf("Did not converge in %d iterations\n",ctl.max_iter_k);
  else
		if (err_flag > 0)printf ( "Converged in %d iterations\n", err_flag);
		else printf("err_flag = %d\n",err_flag);
  if (flx_tran ==1) err_flag = tran_loop();
  return(0);
}
/*------------------------------------------------------------------------*/
/*------------INPUT CELL LIBRARY DATA -----------------------------------*/
/* data loading from cell cross section file */
int cell_input(char *filename)
{
  char buf[100];
  FILE *fp;
  int i, j, p, q;

  /* Open an existing file for reading. Fail if file doesn't exist. */
  if( (fp = fopen( filename, "r")) != NULL )
  {
	 /* Read data from file  */
	 fgets (buf, 80, fp);  /* Title card */
	 fscanf( fp, "%d", &G ); /* number of energy groups */
	 fgets (buf, 100, fp); /* read comments */
	 v = (float *) malloc (G * sizeof ( float) );
	 for (p=0;p<G;p++)
	 {
		fscanf (fp,"%f ",&v[p]);
	 }
	 fgets (buf, 100, fp); /* read comments */
	 fscanf( fp, "%d", &num_cel );  /* Number of cells in the list */
	 fgets (buf, 100, fp); /* read comments */
	 fgets (buf, 100, fp); /* read label card */

	 /* allocate memory for the structure vector */
	 cel = (CELLIB *) malloc (num_cel * sizeof ( CELLIB) );
	 for(i=0;i<num_cel;i++)
	 {
		cel[i].sigma_a = (float *) malloc (G * sizeof (float));
		cel[i].sigma_f = (float *) malloc (G * sizeof (float));
		cel[i].nusigma_f = (float *) malloc (G * sizeof (float));
		cel[i].chi = (float *) malloc (G * sizeof (float));
		cel[i].S_ext = (float *) malloc (G * sizeof (float));
		cel[i].sigma_tr = (float *) malloc (G * sizeof (float));
		cel[i].sigma_r = (float *) malloc (G * sizeof (float));
		cel[i].sigma_s = (float **) malloc (G * sizeof (float));
		for(j=0;j<G;j++)
		{
		  cel[i].sigma_s[j] = (float *) malloc (G * sizeof (float));
		}
	 }
	 /* read cell id */
	 for (i = 0; i < num_cel; i++)
	 {
		fscanf( fp, "%s", &cel[i].name);
		fgets (buf, 80, fp); /* read comments */

		for(p=0;p<G;p++)
		{
		  fscanf(fp, "%f ", &cel[i].sigma_a[p]);
		}
		fgets (buf, 100, fp); /* read comments */
		for(p=0;p<G;p++)
		{
		  fscanf(fp, "%f ", &cel[i].sigma_f[p]);
		}
		fgets (buf, 100, fp); /* read comments */
		for(p=0;p<G;p++)
		{
		  fscanf(fp, "%f ", &cel[i].nusigma_f[p]);
		}
		fgets (buf, 100, fp); /* read comments */
		for(p=0;p<G;p++)
		{
		  fscanf(fp, "%f ", &cel[i].sigma_tr[p]);
		}
		fgets (buf, 100, fp); /* read comments */
		for(p=0;p<G;p++)
		{
		  fscanf(fp, "%f ", &cel[i].sigma_r[p]);
		}
		fgets (buf, 100, fp); /* read comments */

		for(p=0;p<G;p++)
		{
		  for(q=0;q<G;q++)
		  {
			 fscanf(fp, "%f ", &cel[i].sigma_s[p][q]);
		  }
		  fgets (buf, 100, fp); /* read comments */
		}
		for(p=0;p<G;p++)
		{
		  fscanf(fp, "%f ", &cel[i].chi[p]);
		}
		fgets (buf, 100, fp); /* read comments */
		for(p=0;p<G;p++)
		{
		  fscanf(fp, "%f ", &cel[i].S_ext[p]);
		}
		fgets (buf, 100, fp); /* read comments */

	 }
	 fclose( fp );
	 return (1); /* set flag to indicate data is setup */
  }
  else	/* File did not exist. Print message */
  {
	 post_message("Could not open the cell definition file");
	 return(0);
  }
}
/*-----------------------------------------------------------------------*/
/*----------------INPUT CONFIGURATION AND GRID INFO ---------------------*/

/* data loading from grid file */
int grid_input(char *filename)
{
	char buf[100];
	FILE *fp;
	int i, j, k, p, q, cel_id;

	if (cel_set == 0)
	{
		post_message("Cells must be defined first");
		return(0);
	}

	/* Open an existing file for reading. Fail if file doesn't exist. */
	if( (fp = fopen( filename, "r")) != NULL )
	{
		/* Read data from file  */
		/* Example format:
		Title card
		snap.his   <--- output filename for iteration / time history
		map.fou    <--- output filename for flux map at print intervals
		//Control parameters steady state calc's
		100        <--- Maximum # of flux iterations
		.001       <--- Tolerance on the flux
		100        <--- Maximum # of criticality iterations
		.00001     <--- Tolerance on the multiplication constant
		1.2        <--- w parameter for SOR
		1.00       <--- k: initial multiplication constant
		1          <--- initial flux profile type: 1=flat  2=cosine (n/a)
		1          <--- Normalize the flux: 1=yes  0=no
		//Control parameters for transient calc's
		0 			  <--- Perform transient calculations: 0=no 1=yes
		0.0		  <--- Start time
		1.0        <--- Finish time
		.001 0.000001 .1  <--- Nominal time step, min time step, max time step
		//Reactor Grid Definition
		9 rows
		6 columns
		//     	1         2         3         4         5         6         7
		//34567890123456789012345678901234567890123456789012345678901234567890
			  10.  10.  10.  10.  10.  10.  <- grid sizes
		10.  MOD  MOD  MOD  MOD  MOD  MOD  <- row 1
		10.  MOD  MOD  MOD  MOD  MOD  MOD  <- row 2
		10.  MOD  MOD  MOD  MOD  MOD  MOD  <- row 3
		10.  MOD  MOD  FUEL FUEL MOD  MOD  <- row 4
		10.  MOD  MOD  FUEL FUEL MOD  MOD  <- row 5
		10.  MOD  MOD  FUEL FUEL MOD  MOD  <- row 6
		10.  MOD  MOD  MOD  MOD  MOD  MOD  <- row 7
		10.  MOD  MOD  MOD  MOD  MOD  MOD  <- row 8
		10.  MOD  MOD  MOD  MOD  MOD  MOD  <- row 9
		*/
		fgets (buf, 100, fp); //title card
		fscanf (fp, "%s", &hist_filename);
		fgets (buf, 100, fp);
		fscanf (fp, "%s", &out_filename);
		fgets (buf, 100, fp);
		//Control parameters steady state calc's
		fgets (buf, 100, fp); //ss comment card
		fscanf (fp, "%d", &ctl.max_iter_flux);
		fgets (buf, 100, fp); /* read comments */
		fscanf (fp, "%f", &ctl.tolerance_flux);
		fgets (buf, 100, fp);
		fscanf (fp, "%d", &ctl.max_iter_k);
		fgets (buf, 100, fp);
		fscanf (fp, "%f", &ctl.tolerance_k);
		fgets (buf, 100, fp);
		fscanf (fp, "%f", &ctl.w);
		fgets (buf, 100, fp);
		fscanf (fp, "%f", &ctl.k1);
		fgets (buf, 100, fp);
		fscanf( fp, "%f", &flx_profile);
		fgets (buf, 80, fp);
		fscanf( fp, "%d", &flx_norm);
		fgets (buf, 80, fp);
		//Control parameters for transient calc's
		fgets (buf, 80, fp);
		fscanf( fp, "%d", &flx_tran); //transient flag
		fgets (buf, 80, fp);
		fscanf( fp, "%f %f %f", &t_init, &t_fin, &dt_print);
		fgets (buf, 80, fp);
		fscanf( fp, "%f %f %f", &dt_init, &dt_min, &dt_max);
		fgets (buf, 80, fp);
		fscanf( fp, "%f %f", &ctl.a, &ctl.b); //controller gains
		fgets (buf, 80, fp);
		//Reactor Grid Definition
		fgets (buf, 100, fp);
		/* Number of columns in the reactor fuel cell array */
		fscanf( fp, "%d", &num_col );
		fgets (buf, 100, fp);
		/* Number of rows in the reactor fuel cell array */
		fscanf( fp, "%d", &num_row );
		fgets (buf, 100, fp);
		fscanf( fp, "%f", &del_ext); /* flux extrapolation length */
		fgets (buf, 100, fp);
		/* Make allowance for phantom grid points */
		max_row = num_row+2;
		max_col = num_col+2;

		/* 2 label cards */
		fgets (buf, 100, fp); /* read comment */
		fgets (buf, 100, fp); /* read comment */

		/* allocate memory for the grid structure array:
				grd[numrows+2][numcols+2]  	ie grd[i][j] or grd[y][x]  */
		grd = (GRDLIB **) malloc((unsigned) max_row *sizeof(GRDLIB));
		if (!grd)
			{
				post_message("grid memory allocation failure 1");
				return(0);
			}
		for(i=0;i<max_row;i++)
		{
			grd[i]=(GRDLIB *) malloc((unsigned) max_col * sizeof(GRDLIB));
			if (!grd[i])
			{
				post_message("grid memory allocation failure 2");
				return(0);
			}
		}
		for(i=0;i<max_row;i++)
		{
		  for(j=0;j<max_col;j++)
		  {
				grd[i][j].sigma_a = (float *) malloc (G * sizeof (float));
				grd[i][j].nusigma_f = (float *) malloc (G * sizeof (float));
				grd[i][j].chi = (float *) malloc (G * sizeof (float));
				grd[i][j].S_ext = (float *) malloc (G * sizeof (float));
				grd[i][j].sigma_r = (float *) malloc (G * sizeof (float));
				grd[i][j].dif = (float *) malloc (G * sizeof (float));
				grd[i][j].flux = (float *) malloc (G * sizeof (float));
				grd[i][j].flux_old = (float *) malloc (G * sizeof (float));
				grd[i][j].sigma_s = (float **) malloc (G * sizeof (float));
				for(k=0;k<G;k++)
				{
					grd[i][j].sigma_s[k] = (float *) malloc (G * sizeof (float));
				}
		  }
		}


		/* set up dx and dy vectors */
		dx = (float *) malloc (max_col * sizeof (float));
		dy = (float *) malloc (max_row * sizeof (float));

		for (j = 1; j <= num_col; j++)   fscanf( fp, "%f", &dx[j]);
		fgets (buf, 100, fp); /* read comments */
		dx[0] = del_ext *2.0;
		dx[num_col+1] = dx[0];
		dy[0] = del_ext *2.0;
		dy[num_row+1] = dy[0];

		/* read cell names for each grid point */
		for (i = 1; i <= num_row; i++)
		{
			fscanf( fp, "%f", &dy[i]);
			for (j = 1; j <= num_col; j++)
			{
				fscanf( fp, "%s", &grd[i][j].name);
				cel_id = -1;
				for (k = 0; k < num_cel; k++)
				{
					if (!strcmp (cel[k].name, grd[i][j].name) )
					{
						cel_id = k;
						break;
					}
				}
				if (cel_id < 0 )
				{
					printf ("Grid %d %d name is unknown", i,j);
					getch();
					return(0);
				}
				/* assign group constants to grid cells */
				for(p=0;p<G;p++)
				{
					grd[i][j].sigma_a[p] = cel[cel_id].sigma_a[p];
					grd[i][j].nusigma_f[p] = cel[cel_id].nusigma_f[p];
					grd[i][j].sigma_r[p] = cel[cel_id].sigma_r[p];
					grd[i][j].chi[p] = cel[cel_id].chi[p];
					grd[i][j].S_ext[p] = cel[cel_id].S_ext[p]/dx[j]/dy[i];
					for(q=0;q<G;q++)
					{
						grd[i][j].sigma_s[p][q] = cel[cel_id].sigma_s[p][q];
					}
					grd[i][j].dif[p] = 1/(3*cel[cel_id].sigma_tr[p]);
				}
			}
			fgets (buf, 100, fp); /* read comment */
		}
		/* set up diffusion coefficients for the phantom cells */
		for (p=0;p<G;p++)
		{
			for (j=1;j<=num_col;j++)
			{
				grd[0][j].dif[p] = grd[1][j].dif[p];
				grd[num_row+1][j].dif[p] = grd[num_row][j].dif[p];
			}
			for (i=1;i<=num_row;i++)
			{
				grd[i][0].dif[p] = grd[i][1].dif[p];
				grd[i][num_col+1].dif[p] = grd[i][num_col].dif[p];
			}
			grd[0][0].dif[p] = 0.0;
			grd[0][num_col+1].dif[p] = 0.0;
			grd[num_row+1][0].dif[p] = 0.0;
			grd[num_row+1][num_col+1].dif[p] = 0.0;
		}
		fclose( fp );
		return (1); /* set flag to indicate data is setup */
	}
	else	/* File did not exist. Print message */
	{
		post_message("Could not open the configuration input file");
		return(0);
	}

}
/*-----------------------------------------------------------------------*/
/*------------------SET FLUX PROFILE ------------------------------------*/

/* set initial flux distribution */
int flux_input(float flx_profile)
{
	int i, j, p;

	if (grd_set == 0)
	{
		post_message("Grid must be defined first");
		return(0);
	}
	for(p=0;p<G;p++)
	{
		for (i =1; i <= num_row; i++)
		{
			for (j = 1; j <= num_col; j++)
			{
				if (flx_profile >= 0.0) grd[i][j].flux[p]=flx_profile;
				else grd[i][j].flux[p]=1.0; // add cosine later ???
			}
		}
	}
	for (p=0;p<G;p++)
	{
		for (j=1;j<=num_col;j++)
		{
			grd[0][j].flux[p] = 0.0;
			grd[num_row+1][j].flux[p] = 0.0;
		}
		for (i=1;i<=num_row;i++)
		{
			grd[i][0].flux[p] = 0.0;
			grd[i][num_col+1].flux[p] = 0.0;
		}
		grd[0][0].flux[p] = 0.0;
		grd[0][num_col+1].flux[p] = 0.0;
		grd[num_row+1][0].flux[p] = 0.0;
		grd[num_row+1][num_col+1].flux[p] = 0.0;
	}
	return (1); /* set flag to indicate data is setup */
}
/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/
/* ---------MAIN SOR LOOP - modified to minimize storage -----------------*/
int SORM_loop (float w)
{
	int i,j,n, stored, iterk = 0;
	float xmax, error, errork, Sn1, Sn2;
	FILE *fp;

	if (flx_set == 0)
	{
		post_message("Flux must be initialized first");
		return(0);
	}
	if (( fp = fopen(hist_filename,"w")) == NULL)
	{
		printf("Could not open %s",hist_filename);
		return(0);
	}
	fprintf(fp, "Steady state results\n");
	fprintf(fp, "iteration    k    Sn\n");

	ctl.k2 = ctl.k1;
	while ( 1)
	{
		xmax = 0.0;
		errork = 0.0;
		/* calculate individual group fluxes */
		for(n=0;n<G;n++)
		{
		 error = SORM_step (w,n);
		 stored = storefluxm();
		}
		/* calculate new source term and multiplication constant */
		Sn1 = 0.0; Sn2 = 0.0;
		for(i=1;i<=num_row;i++)
		{
			for(j=1;j<=num_col;j++)
			{
				for (n=0;n<G;n++)
				{
					Sn1 = Sn1 + grd[i][j].nusigma_f[n]*grd[i][j].flux_old[n]
									+grd[i][j].S_ext[n];
					Sn2 = Sn2 + grd[i][j].nusigma_f[n]*grd[i][j].flux[n]
									+grd[i][j].S_ext[n];
				}
			}
		}
		ctl.k2 = ctl.k1;
		ctl.k1 = ctl.k1*Sn2/Sn1;
		iterk++;
		errork = fabs( ctl.k1-ctl.k2 );

		/* clip and normalize the flux */
		for ( i = 1; i <= num_row; i++)
		{
			for ( j = 1; j <= num_col; j++)
			{
				for(n=0;n<G;n++)
				{
					if (grd[i][j].flux[n] < 0.0) grd[i][j].flux[n] = 0.0;
					if (grd[i][j].flux[n] > xmax) xmax = grd[i][j].flux[n];
				}
			}
		}
		if (xmax <= 0.0 ) xmax = 1.0;
		if (flx_norm == 1)
		{
			for ( i = 1; i <= num_row; i++)
			{
				for ( j = 1; j <= num_col; j++)
				{
					for(n=0;n<G;n++)
					{
						grd[i][j].flux[n] = grd[i][j].flux[n] / xmax;
						grd[i][j].flux_old[n] = grd[i][j].flux[n];
					}
				}
			}
		}
		printf ("iteration %d,  k = %f %f, Sn= %f %f\n",
					 iterk, ctl.k1, ctl.k2, Sn1, Sn2);
		fprintf (fp,"%d %f %f %f %f\n",
					 iterk, ctl.k1, ctl.k2, Sn1, Sn2);
		 /* check convergence for k */
		if ( errork < ctl.tolerance_k && error < ctl.tolerance_flux )
		{
		stored = storefluxm();   /* Store final flux */
		fclose(fp);
		return (iterk);
		}

		if ( iterk >= ctl.max_iter_k)
		{
		stored = storefluxm();   /* Store final flux */
		fclose(fp);
		return (-1);
		}
	}
}
/* -----------------------------------------------------------------------*/

/* -----------INNER LOOP FOR SOR - modified to minimize storage ----------*/
float SORM_step (float w, int g)
{
	int i, j;
	float error = 0.0;
	float a_p, a_n, a_s, a_w, a_e, b_p;
	float del_ew, del_ns, del_n, del_s, del_e, del_w;
	float D_nf, D_sf, D_ef, D_wf;
	int north, south, east, west, z;
	/*                                    x or j-->
		 N              (i-1)         0  1  2  3  4  5  6  7
		 |                |           1                 ^-num_col
	W-- P-- E   (j-1)--(i,j)--(j+1)  2
		 |                |           3
		 S              (i+1)       y 4        typical grid point property:
											or 5            grd[i][j].dif[n]
											 i 6
											 | 7       (7,3) <--- i=7, j=3
											 | 8
												9 <-num_row
												10
	 For grid of cells of size num_rows x num_col (9x6 in this example), there
	 are extra outside rows and columns used to store grid boundary data.  It is
	 assumed that the extrapolated flux extends into these extra regions for
	 numerical purposes.

	*/
	for ( i = 1; i <= num_row; i++)
	{
		for ( j = 1; j <= num_col; j++)
		{
			north = i - 1;   /* get locations of neighbours */
			south = i + 1;
			east = j + 1;
			west = j - 1;
			if (num_row == 1) //If 1 row there is no diffusion north-south
			{
				grd[north][j].flux[g] = grd[i][j].flux[g];
				grd[south][j].flux[g] = grd[i][j].flux[g];
			}
			if (num_col == 1) //If 1 col there is no diffusion east-west
			{
				grd[i][east].flux[g] = grd[i][j].flux[g];
				grd[i][west].flux[g] = grd[i][j].flux[g];
			}
			del_n = (dy[north]+dy[i])/2.0;/* delta distance from P to neighbours*/
			del_s = (dy[south]+dy[i])/2.0;
			del_e = (dx[east]+dx[j])/2.0;
			del_w = (dx[west]+dx[j])/2.0;
			del_ns = dy[i]; /* width of cell in north-south direction */
			del_ew = dx[j]; /* width of cell in east-west direction */
			/* Average diffusion coefficients estimated at the cell faces */
			D_nf = (grd[north][j].dif[g] + grd[i][j].dif[g])/2.0;
			D_sf = (grd[south][j].dif[g] + grd[i][j].dif[g])/2.0;
			D_ef = (grd[i][east].dif[g] + grd[i][j].dif[g])/2.0;
			D_wf = (grd[i][west].dif[g] + grd[i][j].dif[g])/2.0;

			a_p =(D_nf/del_n+D_sf/del_s)/del_ns + (D_ef/del_e+D_wf/del_w)/del_ew
					+ grd[i][j].sigma_r[g];
			a_n = - D_nf/del_n * grd[north][j].flux[g]/del_ns;
			a_s = - D_sf/del_s * grd[south][j].flux[g]/del_ns;
			a_e = - D_ef/del_e * grd[i][east].flux[g]/del_ew;
			a_w = - D_wf/del_w * grd[i][west].flux[g]/del_ew;

			b_p  = 0.0;
			for(z=0;z<G;z++)
			{
				b_p = b_p + (grd[i][j].sigma_s[z][g] +
								(1/ctl.k1)*grd[i][j].chi[g]*grd[i][j].nusigma_f[z])
								*grd[i][j].flux[z];
			}
			b_p = b_p - grd[i][j].sigma_s[g][g]* grd[i][j].flux[g]
							+ grd[i][j].S_ext[g] ;
			grd[i][j].flux_old[g] = grd[i][j].flux[g];
			grd[i][j].flux[g] = (-a_n - a_s - a_e - a_w +  b_p) / a_p;
			grd[i][j].flux[g] = w * grd[i][j].flux[g] +(1-w)*grd[i][j].flux_old[g];
			error = max(fabs( grd[i][j].flux[g] - grd[i][j].flux_old[g] ), error);
		}
	}
	return error; /* currently no inner convergence criteria on the flux */
}
/* -----------------------------------------------------------------------*/
/* -----------TRANSIENT FLUX CALC - MAIN LOOP ----------------------------*/
int tran_loop()
{
	int i,j,g, stored;
	float error, t=t_init, dt=dt_init, t_print=t_init, flx_new, flx_old;
	FILE *fp;
	flx_setpoint = 1.0;
	if (flx_set == 0)
	{
		post_message("Flux must be initialized first");
		return(0);
	}
	if (( fp = fopen(hist_filename,"a")) == NULL)
	{
		printf("Could not open %s",hist_filename);
		return(0);
	}
	fprintf(fp, "Transient results\n");
	fprintf(fp, "time    k    flux\n");
	while ( t <= t_fin)
	{
		/* calculate individual group fluxes */
		for(g=0;g<G;g++)
		{
		 error = tran_step(g, dt);
		}
		t=t+dt;
		/* adjust dt based on change in flux */

		/* adjust k as per RRS algorithm */
		flx_new = grd[max_row/2][max_col/2].flux[G-1];
		flx_old = grd[max_row/2][max_col/2].flux_old[G-1];
		ctl.k1 = ctl.k1 + ctl.a*(flx_new-flx_setpoint)
			+ ctl.b*(flx_new-flx_old)/dt;
//		if (ctl.k1<1.e-06)ctl.k1=1.e-06;
		printf ("t=%g,  k =%f , flux= %f\n",
					 t, ctl.k1, flx_new);
		fprintf (fp,"%g %f %f\n",t, ctl.k1, flx_new);
		 /* print flux snapshot at intervals */
		if ( t >= t_print)
		{
		stored = storetrans(t);
		t_print = t_print + dt_print;
		}
	}
	fclose (fp);
	return(1);
}
/* -----------------------------------------------------------------------*/
/* -----------TRANSIENT STEP----------------------------------------------*/
float tran_step(int g, float dt)
{
	int i, j;
	float error = 0.0;
	float a_p, a_n, a_s, a_w, a_e, b_p;
	float del_ew, del_ns, del_n, del_s, del_e, del_w;
	float D_nf, D_sf, D_ef, D_wf;
	int north, south, east, west, z;
	/*                                    x or j-->
		 N              (i-1)         0  1  2  3  4  5  6  7
		 |                |           1                 ^-num_col
	W-- P-- E   (j-1)--(i,j)--(j+1)  2
		 |                |           3
		 S              (i+1)       y 4        typical grid point property:
											or 5            grd[i][j].dif[n]
											 i 6
											 | 7       (7,3) <--- i=7, j=3
											 | 8
												9 <-num_row
												10
	 For grid of cells of size num_rows x num_col (9x6 in this example), there
	 are extra outside rows and columns used to store grid boundary data.  It is
	 assumed that the extrapolated flux extends into these extra regions for
	 numerical purposes.

	*/
	for ( i = 1; i <= num_row; i++)
	{
		for ( j = 1; j <= num_col; j++)
		{
			north = i - 1;   /* get locations of neighbours */
			south = i + 1;
			east = j + 1;
			west = j - 1;
			if (num_row == 1) //If 1 row there is no diffusion north-south
			{
				grd[north][j].flux[g] = grd[i][j].flux[g];
				grd[south][j].flux[g] = grd[i][j].flux[g];
			}
			if (num_col == 1) //If 1 col there is no diffusion east-west
			{
				grd[i][east].flux[g] = grd[i][j].flux[g];
				grd[i][west].flux[g] = grd[i][j].flux[g];
			}
			del_n = (dy[north]+dy[i])/2.0;/* delta distance from P to neighbours*/
			del_s = (dy[south]+dy[i])/2.0;
			del_e = (dx[east]+dx[j])/2.0;
			del_w = (dx[west]+dx[j])/2.0;
			del_ns = dy[i]; /* width of cell in north-south direction */
			del_ew = dx[j]; /* width of cell in east-west direction */
			/* Average diffusion coefficients estimated at the cell faces */
			D_nf = (grd[north][j].dif[g] + grd[i][j].dif[g])/2.0;
			D_sf = (grd[south][j].dif[g] + grd[i][j].dif[g])/2.0;
			D_ef = (grd[i][east].dif[g] + grd[i][j].dif[g])/2.0;
			D_wf = (grd[i][west].dif[g] + grd[i][j].dif[g])/2.0;

			a_p =((D_nf/del_n+D_sf/del_s)/del_ns+(D_ef/del_e+D_wf/del_w)/del_ew
					+ grd[i][j].sigma_r[g])*v[g]*dt+1.0;
			a_n = - D_nf/del_n * grd[north][j].flux[g]/del_ns * v[g]*dt;
			a_s = - D_sf/del_s * grd[south][j].flux[g]/del_ns * v[g]*dt;
			a_e = - D_ef/del_e * grd[i][east].flux[g]/del_ew * v[g]*dt;
			a_w = - D_wf/del_w * grd[i][west].flux[g]/del_ew * v[g]*dt;

			b_p  = 0.0;
			for(z=0;z<G;z++)
			{
				b_p = b_p + (grd[i][j].sigma_s[z][g] +
								(1/ctl.k1)*grd[i][j].chi[g]*grd[i][j].nusigma_f[z])
								*grd[i][j].flux[z];
			}
			b_p = (b_p - grd[i][j].sigma_s[g][g]* grd[i][j].flux[g]
							+ grd[i][j].S_ext[g])*v[g]*dt + grd[i][j].flux[g] ;
			grd[i][j].flux_old[g] = grd[i][j].flux[g];
			grd[i][j].flux[g] = (-a_n - a_s - a_e - a_w +  b_p) / a_p;
			error = max(fabs( grd[i][j].flux[g] - grd[i][j].flux_old[g] ), error);
		}
	}
	return error;
}
/* -----------------------------------------------------------------------*/

/* -------- Store Flux in File "Out.put"--------- */

int storefluxm()
{
int i,j,g,k;

FILE *fp;

	if (( fp = fopen(out_filename,"w")) != NULL)
	{
	  fprintf(fp,"Final Flux\n");
	  /* Print the # of neutron flux groups */
	  fprintf( fp, "%d  <-- # of neutron flux groups\n", G);
	  fprintf(fp,"%d %d <-- # columns and rows\n", max_col, max_row);

	  for(g=0;g<G;g++)
	  {
		 fprintf(fp,"\n\n");
		 for ( i = 0; i < max_row; i++)
		 {
			for ( j = 0; j < max_col; j++)
			{
				fprintf(fp,"%5.3f ", grd[i][j].flux[g]);
			}
			fprintf(fp, "\n");
		 }
	  }
	  // print grid material properties to output file
/*     for(k=0;k<G;k++)
	  {
		 fprintf(fp,"\n\n sigma_r nusigma_f dif sigma_s (%d->g)\n", k);
		 for ( i = 1; i <= num_row; i++)
		 {
			for ( j = 1; j <= num_col; j++)
			{
				fprintf(fp,"%f %f %f", grd[i][j].sigma_r[k], grd[i][j].nusigma_f[k],
								 grd[i][j].dif[k]);
				for(g=0; g < G; g++) fprintf(fp,"%f ", grd[i][j].sigma_s[k][g]);
				fprintf(fp,"\n");
			}
			fprintf(fp, "\n");
		 }
	  }
*/
	  fprintf(fp,"dif\n");
	  for(k=0;k<G;k++)
	  {
		 for ( i = 0; i < max_row; i++)
		 {
			for ( j = 0; j < max_col; j++)fprintf(fp,"%f ", grd[i][j].dif[k]);
			fprintf(fp,"\n");
		 }
	  }
	  fprintf(fp,"sigma_r\n");
	  for(k=0;k<G;k++)
	  {
		 for ( i = 1; i <= num_row; i++)
		 {
			for ( j = 1; j <= num_col; j++)fprintf(fp,"%f ", grd[i][j].sigma_r[k]);
			fprintf(fp,"\n");
		 }
	  }
	  fprintf(fp,"nusigma_f\n");
	  for(k=0;k<G;k++)
	  {
		 for ( i = 1; i <= num_row; i++)
		 {
			for ( j = 1; j <= num_col; j++)fprintf(fp,"%f ", grd[i][j].nusigma_f[k]);
			fprintf(fp,"\n");
		 }
	  }
	  fclose(fp);
	  return(1);
	}
	else        /* File did not exist. Print message */
	{
		post_message("Could not open the output file");
		return(0);
	}
}
/* -------- Append transient flux map to output file--------- */

int storetrans(float t)
{
int i,j,g,k;

FILE *fp;

	if (( fp = fopen(out_filename,"a")) != NULL)
	{
	  for(g=0;g<G;g++)
	  {
		 fprintf(fp,"time = %g\n", t);
		 for ( i = 0; i < max_row; i++)
		 {
			for ( j = 0; j < max_col; j++)
			{
				fprintf(fp,"%5.3f ", grd[i][j].flux[g]);
			}
			fprintf(fp, "\n");
		 }
	  }
	  fclose(fp);
	  return(1);
	}
	else        /* File did not exist. Print message */
	{
		post_message("Could not append to the output file");
		return(0);
	}
}
/*--------------------------post message box-------------------------------*/
void post_message ( char *message)
{
  printf("%s",message);
}


